# -*- coding=utf-8 -*-
from __future__ import absolute_import, unicode_literals

from .functools import partialmethod
from .tempfile import NamedTemporaryFile


__all__ = [
    "NamedTemporaryFile",
    "partialmethod"
]
